/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.5 Funktionen Ganglinie
 * Copyright (C) 2011-2015 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.gllib.gllib.domain;

import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Objects;
import java.util.Set;
import java.util.SortedMap;
import java.util.TreeMap;

import de.bsvrz.iav.gllib.gllib.shared.LocalDateTimeKonverter;
import de.bsvrz.sys.funclib.kappich.annotations.NotNull;

/**
 * Für Messquerschnitte angepasste Ganglinie. Die vier Verkehrswerte QKfz, QLkw,
 * VPkw und VLkw werden als Gruppe pro Zeitstempel gesichert. Aus diesen Werten
 * lassen sich die drei davon abhängigen Größe QPkw, VKfz und QB berechnen.
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 */
public class MqGanglinie {

    /**
     * Standardordung der Approximation. Nur für B-Spline relevant: {@value} .
     */
    public static final byte APPROX_STANDARD_ORDNUNG = 5;

    /** Die PID des Messquerschnitts, zu dem die Ganglinie gehört. */
    private final MessQuerschnitt messQuerschnitt;

    /** Parameter für die Berechnung von QB, Standard ist 2.0. */
    private float k1 = 2.0f;

    /** Parameter für die Berechnung von QB, Standard ist 0.01. */
    private float k2 = 0.01f;

    private final long ganglinieId;

    private SortedMap<Long, Messwerte> messwerte = new TreeMap<>();

    /**
     * Zeitpunkt der letzten Verschmelzung, Standard ist der Zeitstempel der
     * Instanzierung.
     */
    private LocalDateTime letzteVerschmelzung = LocalDateTime.now();

    /** Anzahl der Verschmelzung mit anderen Ganglinien, Standard ist 1. */
    private long anzahlVerschmelzungen = 1;

    /**
     * Identifier für das mit der Ganglinie verknüpfte Ereignis, Standard ist null.
     */
    private EreignisTyp ereignisTyp = EreignisTyp.KEIN_EREIGNIS_TYP;

    /**
     * Flag, ob die Ganglinie eine Referenzganglinie darstellt, Standard ist false.
     */
    private boolean referenz;

    /** Typ der Ganglinie, Standard ist {@link #TYP_ABSOLUT}. */
    private GanglinieTyp typ = GanglinieTyp.ABSOLUT;

    /** Art der Approximation, Standard ist {@link ApproximationTyp#BSPLINE}. */
    private ApproximationTyp approximationDaK = ApproximationTyp.BSPLINE;

    /** Ordnung des B-Spline, Standard ist {@link #APPROX_STANDARD_ORDNUNG}. */
    private int bSplineOrdnung = APPROX_STANDARD_ORDNUNG;

    /** Das Intervall für das die Ganglinie prognostiziert wird. */
    private Intervall prognoseZeitraum;

    /** Cache der Einzelganglinie. */
    private DoubleGanglinie gQKfz;

    /** Cache der Einzelganglinie. */
    private DoubleGanglinie gQLkw;

    /** Cache der Einzelganglinie. */
    private DoubleGanglinie gVPkw;

    /** Cache der Einzelganglinie. */
    private DoubleGanglinie gVLkw;

    private boolean approximationAktuell;

    public MqGanglinie(MessQuerschnitt messQuerschnitt) {
        this(0L, messQuerschnitt, EreignisTyp.KEIN_EREIGNIS_TYP);
    }

    public MqGanglinie(long ganglinieId, MessQuerschnitt messQuerschnitt) {
        this(ganglinieId, messQuerschnitt, EreignisTyp.KEIN_EREIGNIS_TYP);
    }

    public MqGanglinie(MessQuerschnitt messQuerschnitt, EreignisTyp ereignisTyp) {
        this(0L, messQuerschnitt, ereignisTyp);
    }

    public MqGanglinie(long ganglinieId, MessQuerschnitt messQuerschnitt, EreignisTyp ereignisTyp) {
        this.ganglinieId = ganglinieId;
        this.messQuerschnitt = messQuerschnitt;
        this.ereignisTyp = ereignisTyp;
    }

    public MqGanglinie erzeugeKopie() {
        return erzeugeKopieMitId(ganglinieId);
    }
    
    public MqGanglinie erzeugeKopieMitId(long id) {

        MqGanglinie g = new MqGanglinie(id, messQuerschnitt, ereignisTyp);
        g.messwerte.putAll(messwerte);
        g.setAnzahlVerschmelzungen(anzahlVerschmelzungen);
        g.setApproximationDaK(approximationDaK);
        g.setBSplineOrdnung(bSplineOrdnung);
        g.setK1(k1);
        g.setK2(k2);
        g.setLetzteVerschmelzung(letzteVerschmelzung);
        g.setPrognoseZeitraum(prognoseZeitraum);
        g.setReferenz(referenz);
        g.setTyp(typ);

        return g;
    }

    public MqGanglinie erzeugeKopieMitEreignisTyp(EreignisTyp srcTyp) {

        MqGanglinie g = new MqGanglinie(ganglinieId, messQuerschnitt, srcTyp);
        g.messwerte.putAll(messwerte);
        g.setAnzahlVerschmelzungen(anzahlVerschmelzungen);
        g.setApproximationDaK(approximationDaK);
        g.setBSplineOrdnung(bSplineOrdnung);
        g.setK1(k1);
        g.setK2(k2);
        g.setLetzteVerschmelzung(letzteVerschmelzung);
        g.setPrognoseZeitraum(prognoseZeitraum);
        g.setReferenz(referenz);
        g.setTyp(typ);

        return g;
    }

    
    /**
     * Gibt die Approximation zur Datenverteilerapproximation zurück.
     *
     * @return die Approximation.
     * @see #approximationDaK
     */
    private Approximation<Double> erzeugeApproximation() {
        switch (approximationDaK) {
        case BSPLINE:
            return new BSpline(bSplineOrdnung);
        case CUBICSPLINE:
            return new CubicSpline();
        case POLYLINE:
            return new Polyline();
        default:
            return new BSpline((byte) 5);
        }
    }

    /**
     * Gibt die Anzahl der bisherigen Verschmelzungen beim automatischen Lernen
     * zurück.
     *
     * @return Anzahl bisheriger Verschmelzungen
     */
    public long getAnzahlVerschmelzungen() {
        return anzahlVerschmelzungen;
    }

    public ApproximationTyp getApproximationDaK() {
        return approximationDaK;
    }

    /**
     * Gibt die Ordnung des B-Spline zurück. Wird zur Approximation kein B-Spline
     * benutzt, wird der Wert ignoriert.
     *
     * @return die Ordnung des B-Spline.
     */
    public int getBSplineOrdnung() {
        return bSplineOrdnung;
    }

    /**
     * Gibt die ID des Ereignistyps der Ganglinie zurück.
     *
     * @return die ID des Ereignistyps.
     */
    public EreignisTyp getEreignisTyp() {
        return ereignisTyp;
    }

    public void setEreignisTyp(EreignisTyp ereignisTyp) {
        this.ereignisTyp = ereignisTyp;
    }

    /**
     * Gibt die Ganglinie für QB zurück.
     *
     * @return eine einfache mathematische Ganglinie mit den Stützstellen von QB.
     */
    public DoubleGanglinie getGanglinieQB() {
        DoubleGanglinie g;

        g = new DoubleGanglinie();
        for (final Map.Entry<Long, Messwerte> e : messwerte.entrySet()) {
            g.aktualisierePunkt(e.getKey(), e.getValue().getQB());
        }
        g.setApproximation(erzeugeApproximation());

        return g;
    }

    /**
     * Gibt die Ganglinie für QKfz zurück.
     *
     * @return eine einfache mathematische Ganglinie mit den Stützstellen von QKfz.
     */
    public DoubleGanglinie getGanglinieQKfz() {
        DoubleGanglinie g;

        g = new DoubleGanglinie();
        for (final Map.Entry<Long, Messwerte> e : messwerte.entrySet()) {
            g.aktualisierePunkt(e.getKey(), e.getValue().getQKfz());
        }
        g.setApproximation(erzeugeApproximation());

        return g;
    }

    /**
     * Gibt die Ganglinie für QLkw zurück.
     *
     * @return eine einfache mathematische Ganglinie mit den Stützstellen von QLkw.
     */
    public DoubleGanglinie getGanglinieQLkw() {
        DoubleGanglinie g;

        g = new DoubleGanglinie();
        for (final Map.Entry<Long, Messwerte> e : messwerte.entrySet()) {
            g.aktualisierePunkt(e.getKey(), e.getValue().getQLkw());
        }
        g.setApproximation(erzeugeApproximation());

        return g;
    }

    /**
     * Gibt die Ganglinie für QPkw zurück.
     *
     * @return eine einfache mathematische Ganglinie mit den Stützstellen von QPkw.
     */
    public DoubleGanglinie getGanglinieQPkw() {
        DoubleGanglinie g;

        g = new DoubleGanglinie();
        for (final Map.Entry<Long, Messwerte> e : messwerte.entrySet()) {
            g.aktualisierePunkt(e.getKey(), e.getValue().getQPkw());
        }
        g.setApproximation(erzeugeApproximation());

        return g;
    }

    /**
     * Gibt die Ganglinie für VKfz zurück.
     *
     * @return eine einfache mathematische Ganglinie mit den Stützstellen von VKfz.
     */
    public DoubleGanglinie getGanglinieVKfz() {
        DoubleGanglinie g;

        g = new DoubleGanglinie();
        for (final Map.Entry<Long, Messwerte> e : messwerte.entrySet()) {
            g.aktualisierePunkt(e.getKey(), e.getValue().getVKfz());
        }
        g.setApproximation(erzeugeApproximation());

        return g;

    }

    /**
     * Gibt die Ganglinie für VLkw zurück.
     *
     * @return eine einfache mathematische Ganglinie mit den Stützstellen von VLkw.
     */
    public DoubleGanglinie getGanglinieVLkw() {
        DoubleGanglinie g;

        g = new DoubleGanglinie();
        for (final Map.Entry<Long, Messwerte> e : messwerte.entrySet()) {
            g.aktualisierePunkt(e.getKey(), e.getValue().getVLkw());
        }
        g.setApproximation(erzeugeApproximation());

        return g;
    }

    /**
     * Gibt die Ganglinie für QPkw zurück.
     *
     * @return eine einfache mathematische Ganglinie mit den Stützstellen von QPkw.
     */
    public DoubleGanglinie getGanglinieVPkw() {
        DoubleGanglinie g;

        g = new DoubleGanglinie();
        for (final Map.Entry<Long, Messwerte> e : messwerte.entrySet()) {
            g.aktualisierePunkt(e.getKey(), e.getValue().getVPkw());
        }
        g.setApproximation(erzeugeApproximation());

        return g;
    }

    public Intervall getIntervall() {
        if (prognoseZeitraum != null) {
            return prognoseZeitraum;
        }

        if (messwerte.isEmpty()) {
            return null;
        }
        return Intervall.of(messwerte.firstKey(), messwerte.lastKey());
    }

    /**
     * Gibt einen Parameter für die Berechnung von QB zurück.
     *
     * @return der Parameter.
     */
    public float getK1() {
        return k1;
    }

    /**
     * Gibt einen Parameter für die Berechnung von QB zurück.
     *
     * @return der Parameter
     */
    public float getK2() {
        return k2;
    }

    /**
     * Gibt den Zeitpunkt der letzten Verschmelzung als Zeitstempel zurück.
     *
     * @return Zeitstempel
     */
    public LocalDateTime getLetzteVerschmelzung() {
        return letzteVerschmelzung;
    }

    public MessQuerschnitt getMessQuerschnitt() {
        return messQuerschnitt;
    }

    /**
     * Gibt das Prognoseintervall der Ganglinie zurück.
     *
     * @return das Prognoseintervall.
     */
    public Intervall getPrognoseIntervall() {
        return prognoseZeitraum;
    }

    
    public Stuetzstelle<Messwerte> getStuetzstelle(LocalDateTime zeitstempel) {
        return getStuetzstelle(LocalDateTimeKonverter.toMillis(zeitstempel));
    }
    
    public Stuetzstelle<Messwerte> getStuetzstelle(long zeitstempel) {

        if ((prognoseZeitraum != null) && !prognoseZeitraum.contains(zeitstempel)) {
            // Zeitstempel liegt nicht innerhalb der Prognoseganglinie
            return new Stuetzstelle<>(zeitstempel, new Messwerte(null, null, null, null, k1, k2));
        }

        if (!approximationAktuell) {
            gQKfz = getGanglinieQKfz();
            gQLkw = getGanglinieQLkw();
            gVPkw = getGanglinieVPkw();
            gVLkw = getGanglinieVLkw();
            approximationAktuell = true;
        }

        return new Stuetzstelle<>(zeitstempel,
                new Messwerte(gQKfz.getStuetzstelle(zeitstempel).getWert(),
                        gQLkw.getStuetzstelle(zeitstempel).getWert(), gVPkw.getStuetzstelle(zeitstempel).getWert(),
                        gVLkw.getStuetzstelle(zeitstempel).getWert(), k1, k2));
    }

    public List<Stuetzstelle<Messwerte>> getStuetzstellen() {
        List<Stuetzstelle<Messwerte>> liste;

        liste = new ArrayList<>();
        for (final Entry<Long, Messwerte> entry : messwerte.entrySet()) {
            if ((prognoseZeitraum != null) && !prognoseZeitraum.contains(entry.getKey())) {
                continue;
            }

            liste.add(new Stuetzstelle<>(entry.getKey(), entry.getValue()));
        }

        return liste;
    }

    /**
     * Gibt berechnete Stützstellen im angegebenen Intervall in der angegebenen
     * Schrittweite zurück.
     * 
     * @param intervall das Intervall für das Stützstellen geliefert werden sollen
     * @return eine Liste mit den ermittelten Stützstellen
     */
    public @NotNull List<Stuetzstelle<Messwerte>> getStuetzstellen(final Intervall intervall) {
        SortedMap<Long, Messwerte> menge;
        List<Stuetzstelle<Messwerte>> liste = new ArrayList<>();

        menge = messwerte.subMap(intervall.getStart(), intervall.getEnd() + 1);
        for (final long t : menge.keySet()) {
            if ((prognoseZeitraum != null) && !prognoseZeitraum.contains(t)) {
                continue;
            }

            liste.add(getStuetzstelle(t));
        }

        return liste;
    }

    /**
     * Gibt berechnete Stützstellen im prognostizierten Intervall in der angegebenen
     * Schrittweite zurück.
     *
     * @param schrittweite die Schrittweite der Stützstellen für die Berechnung.
     * @return die berechneten Stützstellen.
     * @see #setPrognoseZeitraum(Intervall)
     */
    public List<Stuetzstelle<Messwerte>> getStuetzstellen(final long schrittweite) {
        return getStuetzstellen(getPrognoseIntervall() == null ? getIntervall() : getPrognoseIntervall(), schrittweite);
    }

    /**
     * Gibt berechnete Stützstellen in einem bestimmten Intervall in der angegebenen
     * Schrittweite zurück.
     *
     * @param intervall    das Intervall der Ganglinie für das die Stützstellen
     *                     berechnet werden sollen.
     * @param schrittweite die Schrittweite der Stützstellen für die Berechnung.
     * @return die berechneten Stützstellen.
     * @see #setPrognoseZeitraum(Intervall)
     */
    public List<Stuetzstelle<Messwerte>> getStuetzstellen(final Intervall intervall, final long schrittweite) {
        List<Stuetzstelle<Messwerte>> liste;

        liste = new ArrayList<>();
        for (long t = intervall.getStart(); t <= intervall.getEnd(); t += schrittweite) {
            if ((prognoseZeitraum != null) && !prognoseZeitraum.contains(t)) {
                continue;
            }

            liste.add(getStuetzstelle(t));
        }

        return liste;
    }

    public GanglinieTyp getTyp() {
        return typ;
    }

    /**
     * Bestimmt, ob die Ganglinie absolut oder relativ ist. Für eine relative
     * Ganglinie wird nicht zwischen additiv und multiplikativ unterschieden.
     *
     * @return {@code true}, wenn die Ganglinie absolut ist.
     */
    public boolean isAbsolut() {
        return typ == GanglinieTyp.ABSOLUT;
    }

    /**
     * Besitzt die Ganglinie die Auszeichnung als Referenz?
     *
     * @return <code>true</code>, wenn diese Ganglinie eine Referenzganglinie ist,
     *         sonst <code>false</code>
     */
    public boolean isReferenz() {
        return referenz;
    }

    /**
     * Legt die Anzahl der bisherigen Verschmelzungen fest.
     * <p>
     * <em>Hinweis:</em> Diese Methode ist nicht Teil der öffentlichen API und
     * sollte nicht außerhalb der Ganglinie-API verwendet werden.
     *
     * @param anzahlVerschmelzungen Anzahl der Verschmelzungen
     */
    public void setAnzahlVerschmelzungen(final long anzahlVerschmelzungen) {
        this.anzahlVerschmelzungen = anzahlVerschmelzungen;
    }

    /**
     * Legt die zu verwendende Approximation mit Hilfe einer Datenkatalogkonstante
     * fest.
     *
     * @param approximationDaK der gewünschte Approximationstyp
     */
    public void setApproximationDaK(final ApproximationTyp approximationDaK) {
        this.approximationDaK = approximationDaK;
    }

    /**
     * Legt die Ordnung des B-Spline fest. Wird zur Approximation kein B-Spline
     * benutzt, wird der Wert ignoriert.
     *
     * @param ordnung die neue Ordnung des B-Spline.
     */
    public void setBSplineOrdnung(int ordnung) {
        this.bSplineOrdnung = ordnung;
    }

    /**
     * Legt den Parameter k1 für die Berechnung von QB fest.
     *
     * @param k1 der parameter k1
     */
    public void setK1(final float k1) {
        final SortedMap<Long, Messwerte> neu;

        this.k1 = k1;

        neu = new TreeMap<>();
        for (final Map.Entry<Long, Messwerte> e : messwerte.entrySet()) {
            neu.put(e.getKey(), new Messwerte(e.getValue().getQKfz(), e.getValue().getQLkw(), e.getValue().getVPkw(),
                    e.getValue().getVLkw(), k1, k2));
        }
        messwerte.clear();
        messwerte.putAll(neu);
    }

    /**
     * Legt den Parameter k2 für die Berechnung von QB fest.
     *
     * @param k2 der parameter k2
     */
    public void setK2(final float k2) {
        final SortedMap<Long, Messwerte> neu;

        this.k2 = k2;

        neu = new TreeMap<>();
        for (final Map.Entry<Long, Messwerte> e : messwerte.entrySet()) {
            neu.put(e.getKey(), new Messwerte(e.getValue().getQKfz(), e.getValue().getQLkw(), e.getValue().getVPkw(),
                    e.getValue().getVLkw(), k1, k2));
        }
        messwerte.clear();
        messwerte.putAll(neu);
    }

    /**
     * Legt die Anzahl der bisherigen Verschmelzungen fest.
     * <p>
     * <em>Hinweis:</em> Diese Methode ist nicht Teil der öffentlichen API und
     * sollte nicht außerhalb der Ganglinie-API verwendet werden.
     *
     * @param letzteVerschmelzung die neue Anzahl der Verschmelzungen.
     */
    public void setLetzteVerschmelzung(final LocalDateTime letzteVerschmelzung) {
        this.letzteVerschmelzung = letzteVerschmelzung;
    }

    /**
     * Legt das Prognoseintervall fest. Die getter-Methoden für Stützstellen liefern
     * nur Stützstellen innerhalb dieses Intervalls. Ist das Prognoseintervall
     * gleich {@code null}, dann werden alle vorhanden Stützstellen berücksichtigt.
     *
     * @param prognoseZeitraum ein Intervall.
     */
    public void setPrognoseZeitraum(final Intervall prognoseZeitraum) {
        this.prognoseZeitraum = prognoseZeitraum;
    }

    public void setPrognoseZeitraum(final PrognoseIntervall prognoseZeitraum) {
        long start = LocalDateTimeKonverter.toMillis(prognoseZeitraum.getVon());
        long end = LocalDateTimeKonverter.toMillis(prognoseZeitraum.getBis());
        this.prognoseZeitraum = Intervall.of(start, end);
    }

    /**
     * Kennzeichnet die Ganglinie als Referenzganglinie.
     * <p>
     * <em>Hinweis:</em> Diese Methode ist nicht Teil der öffentlichen API und
     * sollte nicht außerhalb der Ganglinie-API verwendet werden.
     *
     * @param referenz <code>true</code>, wenn diese Ganglinie eine
     *                 Referenzganglinie sein soll, sonst <code>false</code>
     */
    public void setReferenz(final boolean referenz) {
        this.referenz = referenz;
    }

    /**
     * Legt den Ganglinientyp fest.
     * <p>
     * <em>Hinweis:</em> Diese Methode ist nicht Teil der öffentlichen API und
     * sollte nicht außerhalb der Ganglinie-API verwendet werden.
     *
     * @param typ der Typ der Ganglinie.
     */
    public void setTyp(final GanglinieTyp typ) {
        this.typ = typ;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String toString() {
        String result;

        result = getClass().getSimpleName() + "[\n";
        result += GanglinienMQOperationen.formatierterText(this);
        result += "]";

        return result;
    }

    public void aktualisiereMesswerte(Map<Long, Messwerte> neueMesswerte) {
        messwerte.putAll(neueMesswerte);
        approximationAktuell = false;
    }

    public void loescheMesswerte() {
        messwerte.clear();
        approximationAktuell = false;
    }

    public void loescheMesswerte(Set<Long> timestamps) {
        timestamps.forEach((t) -> messwerte.remove(t));
        approximationAktuell = false;
    }

    public boolean hatMesswerte() {
        return !messwerte.isEmpty();
    }

    public void aktualisiereMesswert(long zeitstempel, Messwerte messwert) {
        messwerte.put(zeitstempel, messwert);
        approximationAktuell = false;
    }

    public SortedMap<Long, Messwerte> getMesswerte() {
        return Collections.unmodifiableSortedMap(messwerte);
    }

    public int getStuetzStellenAnzahl() {
        return messwerte.size();
    }

    public boolean hatMesswertFuer(long zeitpunkt) {
        return messwerte.containsKey(zeitpunkt);
    }

    public boolean vergleiche(MqGanglinie other) {
        return Objects.equals(messwerte, other.messwerte);
    }

    public long getGanglinieId() {
        return ganglinieId;
    }

    @Override
    public int hashCode() {
        return Objects.hash(anzahlVerschmelzungen, approximationAktuell, approximationDaK, bSplineOrdnung, ereignisTyp,
                gQKfz, gQLkw, gVLkw, gVPkw, ganglinieId, k1, k2, letzteVerschmelzung, messQuerschnitt, messwerte,
                prognoseZeitraum, referenz, typ);
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null) {
            return false;
        }
        if (!(obj instanceof MqGanglinie)) {
            return false;
        }
        MqGanglinie other = (MqGanglinie) obj;
        return anzahlVerschmelzungen == other.anzahlVerschmelzungen
                && approximationAktuell == other.approximationAktuell && approximationDaK == other.approximationDaK
                && bSplineOrdnung == other.bSplineOrdnung && Objects.equals(ereignisTyp, other.ereignisTyp)
                && Objects.equals(gQKfz, other.gQKfz) && Objects.equals(gQLkw, other.gQLkw)
                && Objects.equals(gVLkw, other.gVLkw) && Objects.equals(gVPkw, other.gVPkw)
                && ganglinieId == other.ganglinieId && Float.floatToIntBits(k1) == Float.floatToIntBits(other.k1)
                && Float.floatToIntBits(k2) == Float.floatToIntBits(other.k2)
                && Objects.equals(letzteVerschmelzung, other.letzteVerschmelzung)
                && Objects.equals(messQuerschnitt, other.messQuerschnitt) && Objects.equals(messwerte, other.messwerte)
                && Objects.equals(prognoseZeitraum, other.prognoseZeitraum) && referenz == other.referenz
                && typ == other.typ;
    }
    
    
}
