/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.5 Funktionen Ganglinie
 * Copyright (C) 2011-2015 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.gllib.gllib.domain;

import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Objects;
import java.util.Set;
import java.util.SortedMap;
import java.util.TreeMap;

/**
 * Repräsentiert eine allgemeine Ganglinie, bestehend aus einer sortierten Menge
 * von Stützstellen und der Angabe eines Interpolationsverfahren.
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 */
public class DoubleGanglinie {

	/** Verfahren zur Berechnung der Punkte zwischen den Stützstellen. */
	private Approximation<Double> approximation;

	private SortedMap<Long, Double> punkte = new TreeMap<>();

	/** Flag, ob die Approximation aktualisiert werden muss. */
	private boolean approximationAktuell;

	/** Cached die Teilintervalle der Ganglinie mit definierten Stützstellen. */
	private final List<Intervall> intervalle;

	/**
	 * Konstruiert eine Ganglinie ohne Stützstellen.
	 */
	public DoubleGanglinie() {
		intervalle = new ArrayList<>();
		setApproximationAktuell(false);
	}

	/**
	 * Markiert zusätzlich die Approximation als nicht mehr aktuell.
	 */
	public void loeschen() {
		setApproximationAktuell(false);
		punkte.clear();
	}

	/**
	 * Kopiert die Stützstellen und das Approximationsverfahren. Der Wert für
	 * {@code approximationAktuell} wird auf false gesetzt.
	 * 
	 * @return die erzeugte Kopie der Ganglinie
	 */
	public DoubleGanglinie erzeugeKopie() {
		DoubleGanglinie g = new DoubleGanglinie();
		g.punkte.putAll(this.punkte);
		g.setApproximation(approximation);
		return g;
	}

	/**
	 * Gibt die Approximation der Ganglinie zurück. Die Approximation wird falls
	 * nötig vorher initialisiert.
	 *
	 * @return die Approximation.
	 */
	public Approximation<Double> getApproximation() {
		if (approximation != null) {
			aktualisiereApproximation();
		}
		return approximation;
	}

	/**
	 * Gibt das Intervall der Ganglinie zurück. Das Intervall besteht aus der ersten
	 * und letzten Stützstelle. Existieren keine Stützstellen wird {@code null}
	 * zurückgegeben.
	 *
	 * @return das Ganglinienintervall oder {@code null}.
	 */
	public Intervall getIntervall() {
		if (punkte.size() == 0) {
			return null;
		}

		return Intervall.of(punkte.firstKey(), punkte.lastKey());
	}

	/**
	 * Bestimmt die Intervalle in denen die Ganglinie definiert ist.
	 *
	 * @return eine Liste von Intervallen.
	 */
	@SuppressWarnings("null")
	public List<Intervall> getIntervalle() {
		if (isApproximationAktuell()) {
			return intervalle;
		}

		Long start, ende;

		intervalle.clear();
		start = null;
		ende = null;

		for (Entry<Long, Double> entry : punkte.entrySet()) {

			Long punktZeit = entry.getKey();
			if (start == null) {
				// Beginn eines neuen Intervalls
				if (entry.getValue() != null) {
					if (punktZeit.equals(punkte.lastKey())) {
						// Die letzte Stützstelle ist das letzte Intervall
						intervalle.add(Intervall.of(punktZeit, punktZeit));
					} else {
						start = punktZeit;
					}
				}
			} else {
				if (punkte.get(punktZeit) == null) {
					// Definitionslücke gefunden
					if (ende != null) {
						intervalle.add(Intervall.of(start, ende));
						start = null;
						ende = null;
					} else {
						intervalle.add(Intervall.of(start, start));
						start = null;
					}
				} else {
					// Intervall verlängern
					if (punkte.get(punktZeit) != null) {
						ende = punktZeit;
					}
					if (punktZeit.equals(punkte.lastKey())) {
						// Die letzte Stützstelle ist das letzte Intervall
						/*
						 * Die Warnung wegen "potential null pointer access" kann ignoriert werden, weil
						 * in der Schleife der Wert irgend wann vorher gesetzt wird.
						 */
						intervalle.add(Intervall.of(start, ende));
					}
				}
			}
		}

		return intervalle;
	}

	/**
	 * Gibt die Stützstelle zu einem bestimmten Zeitpunkt zurück. Es wird die mit
	 * der Approximation berechnete Stützstelle ausgeliefert. Die Approximation muss
	 * dazu zuvor festgelegt worden sein.
	 *
	 * @param zeitstempel der Zeitstempel zu dem eine Stützstelle gesucht wird.
	 * @return die gesuchte Stützstelle.
	 */
	public Stuetzstelle<Double> getStuetzstelle(final long zeitstempel) {
		if (!isValid(zeitstempel)) {
			// Zeitstempel liegt in einem undefinierten Teilintervall
			return new Stuetzstelle<>(zeitstempel, null);
		}

		if (approximation != null) {
			aktualisiereApproximation();
			return approximation.get(zeitstempel);
		}

		throw new IllegalStateException("Es wurde keine Approximation festgelegt.");
	}

	/**
	 * Gibt ein sortiertes Feld der existierenden Stützstellen zurück.
	 *
	 * @return die nach Zeitstempel sortierten Stützstellen.
	 */
	public List<Stuetzstelle<Double>> getStuetzstellen() {
		List<Stuetzstelle<Double>> liste;

		liste = new ArrayList<>();
		for (final Entry<Long, Double> entry : punkte.entrySet()) {
			liste.add(new Stuetzstelle<>(entry.getKey(), entry.getValue()));
		}

		return liste;
	}

	/**
	 * Gibt die existierenden Stützstellen im angegebenen Intervall zurück.
	 *
	 * @param intervall ein Intervall.
	 * @return die Liste der Stützstellen im Intervall, sortiert nach Zeitstempel.
	 */
	public List<Stuetzstelle<Double>> getStuetzstellen(final Intervall intervall) {
		SortedMap<Long, Double> menge;
		List<Stuetzstelle<Double>> liste;

		menge = punkte.subMap(intervall.getStart(), intervall.getEnd() + 1);
		liste = new ArrayList<>();
		for (final long t : menge.keySet()) {
			liste.add(getStuetzstelle(t));
		}

		return liste;
	}

	/**
	 * Prüft ob ein Teilintervall der Ganglinie vollständig definiert ist, also
	 * keine undefinierten Berreiche enthält.
	 *
	 * @param intervall das zu prüfende Intervall.
	 * @return {@code true}, wenn das Teilintervall der Ganglinie keine
	 *         undefinierten Bereiche enthält.
	 * @see #getIntervalle()
	 */
	public boolean isValid(final Intervall intervall) {
		boolean ok;

		ok = false;
		for (final Intervall i : getIntervalle()) {
			if (i.contains(intervall)) {
				ok = true;
				break;
			}
		}

		return ok;
	}

	/**
	 * Prüft ob ein Zeitstempel im definiterten Bereich der Ganglinie liegt.
	 *
	 * @param zeitstempel der zu prüfender Zeitstempel.
	 * @return {@code true}, wenn der Zeitstempel im definierten Bereich der
	 *         Ganglinie liegt.
	 * @see #getIntervalle()
	 */
	public boolean isValid(final long zeitstempel) {
		boolean ok;

		ok = false;
		for (final Intervall i : getIntervalle()) {
			if (i.contains(zeitstempel)) {
				ok = true;
				break;
			}
		}

		return ok;
	}

	/**
	 * Aktualisiert den übergebenen Punkt oder fügt einen neuen Punkt ein, wenn für
	 * den angegebenen Zeitpunkt kein Punkt existiert. Zusätzlich wird die
	 * Approximation als nicht mehr aktuell markiert.
	 * 
	 * @param key   der Zeitstempel des Punkts
	 * @param value der Wert des Punkts
	 * 
	 * @return <code>null</code>, wenn ein neuer Punkt eingefügt wurde. Anderenfalls
	 *         wird der Wert des vorgherigen Punktes geliefert.
	 */
	public Double aktualisierePunkt(Long key, final Double value) {
		setApproximationAktuell(false);
		return punkte.put(key, value);
	}

	/**
	 * Aktualisiert die übergebenen Punkt oder fügt Punkte ein, wenn für den
	 * jeweiligen Zeitpunkt kein Punkt existiert. Zusätzlich wird die Approximation
	 * als nicht mehr aktuell markiert.
	 * 
	 * @param neuePunkte ein Map mit den zu aktualisierenden Punkten
	 */
	public void aktualisierePunkte(Map<Long, Double> neuePunkte) {
		setApproximationAktuell(false);
		punkte.putAll(neuePunkte);
	}

	/**
	 * Entfernt den Datenpunkt mit dem übergebenen Zeitpunkt und markiert zusätzlich
	 * die Approximation als nicht mehr aktuell.
	 * 
	 * @param zeitpunkt der Zeitpunkt des zu entfernenden Datenpunktes
	 * @return der Wert des entfernten Datenpunkts oder <code>null</code>, wenn
	 *         keiner existiert hat
	 */
	public Double entfernePunkt(long zeitpunkt) {
		setApproximationAktuell(false);
		return punkte.remove(zeitpunkt);
	}

	/**
	 * Legt das Approximationsverfahren fest, mit dem die Werte zwischen den
	 * Stützstellen bestimmt werden soll.
	 *
	 * @param approximation das Approximationsverfahren.
	 */
	public void setApproximation(final Approximation<Double> approximation) {
		setApproximationAktuell(false);
		this.approximation = approximation;
	}

	/**
	 * Nimmt eine Stützstelle in die Ganglinie auf. Existiert zu dem Zeitpunkt
	 * bereits eine, wird diese überschrieben.
	 *
	 * @param s die neue Stuützstelle.
	 * @return {@code true}, wenn die Stützstelle neu angelegt wurde und
	 *         {@code false}, wenn eine vorhandene Stützstelle ersetzt wurde.
	 */
	public boolean setStuetzstelle(final Stuetzstelle<Double> s) {
		return punkte.put(s.getZeitstempel(), s.getWert()) == null;
	}

	/**
	 * Ersetzt die Stützstellen der Ganglinie.
	 *
	 * @param stuetzstellen die neuen Stützstellen
	 */
	public void setStuetzstellen(final Collection<Stuetzstelle<Double>> stuetzstellen) {
		setApproximationAktuell(false);
		punkte.clear();
		for (final Stuetzstelle<Double> s : stuetzstellen) {
			punkte.put(s.getZeitstempel(), s.getWert());
		}
	}

	/**
	 * {@inheritDoc}
	 *
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		final SimpleDateFormat format = new SimpleDateFormat("dd.MM.yyyy HH.mm");

		StringBuilder s = new StringBuilder(getClass().getName() + " , Daten:\n");
		for (final Entry<Long, Double> entry : punkte.entrySet()) {
			s.append(format.format(new Date(entry.getKey())) + " --> " + entry.getValue() + "\n");
		}

		return s.toString();
	}

	/**
	 * Aktualisiert falls nötig die Approximation.
	 */
	private void aktualisiereApproximation() {
		if (!isApproximationAktuell()) {
			getIntervalle();
			approximation.setStuetzstellen(getStuetzstellen());
			approximation.initialisiere();
			setApproximationAktuell(true);
		}
	}

	/**
	 * Gibt {@code false} zurück, wenn die Approximation aktuallisiert werden muss,
	 * weil sich die Ganglinie geändert hat.
	 *
	 * @return {@code true}, wenn Ganglinie und Approximation konform gehen und
	 *         {@code false}, wenn die Approximation aktualisiert werden muss.
	 */
	protected boolean isApproximationAktuell() {
		return approximationAktuell;
	}

	/**
	 * Setzt das Flag, ob die Approximation noch gültig ist oder nicht.
	 *
	 * @param approximationAktuell {@code false}, wenn die Approximation
	 *                             aktualisiert werden muss.
	 */
	protected void setApproximationAktuell(final boolean approximationAktuell) {
		this.approximationAktuell = approximationAktuell;
	}

	public Long ersterZeitpunkt() {
		return punkte.firstKey();
	}

	public Long letzterZeitpunkt() {
		return punkte.lastKey();
	}

	public boolean hatZeitpunkt(long zeitpunkt) {
		return punkte.containsKey(zeitpunkt);
	}

	public Double getWert(long zeitpunkt) {
		return punkte.get(zeitpunkt);
	}

	public Set<Long> getZeitpunkte() {
		return punkte.keySet();
	}

	public SortedMap<Long, Double> getWerte() {
		return punkte;
	}

	public int punkteAnzahl() {
		return punkte.size();
	}

	public boolean vergleiche(DoubleGanglinie other) {
		return Objects.equals(punkte, other.punkte);
	}

	@Override
	public int hashCode() {
		return Objects.hash(approximation, approximationAktuell, intervalle, punkte);
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (obj == null) {
			return false;
		}
		if (!(obj instanceof DoubleGanglinie)) {
			return false;
		}
		DoubleGanglinie other = (DoubleGanglinie) obj;
		return Objects.equals(approximation, other.approximation) && approximationAktuell == other.approximationAktuell
				&& Objects.equals(intervalle, other.intervalle) && Objects.equals(punkte, other.punkte);
	}

}
