/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.5 Funktionen Ganglinie
 * Copyright (C) 2011-2015 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.gllib.gllib.domain;

import java.text.MessageFormat;
import java.time.Duration;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map.Entry;
import java.util.Objects;
import java.util.Queue;
import java.util.SortedMap;
import java.util.SortedSet;
import java.util.TreeMap;
import java.util.TreeSet;

/**
 * Helferklasse mit den Operationen auf Ganglinien, deren Stützstellen
 * Gleitkommazahlen sind.
 *
 * @author BitCtrl Systems GmbH, Schumann
 */
public final class GanglinienOperationen {

    /**
     * Kapselt die Informationen die als Ergebnis aus dem Pattern-Matching
     * hervorgehen.
     */
    public static final class PatternMatchingErgebnis {

        /** Die Ergebnisganglinie. */
        private final DoubleGanglinie ganglinie;

        /** Der Index der ausgewählten Ganglinie in der Vergleichsliste. */
        private final int index;

        /** Der Abstand der Ergebnisganglinie zur Referenzganglinie. */
        private final double abstand;

        /** Der Offset um den die Ergebnisganglinie verschoben wurde. */
        private final long offset;

        /**
         * Erzeugt ein Pattern-Matching-Ergebnis.
         *
         * @param ganglinie die Ergebnisganglinie.
         * @param index     der Index der ausgewählten Ganglinie in der Vergleichsliste.
         * @param abstand   der Abstand der Ergebnisganglinie zur Referenzganglinie.
         * @param offset    der Offset um den die Ergebnisganglinie verschoben wurde.
         */
        public PatternMatchingErgebnis(final DoubleGanglinie ganglinie, final int index, final double abstand,
                final long offset) {
            this.ganglinie = ganglinie;
            this.index = index;
            this.abstand = abstand;
            this.offset = offset;
        }

        /**
         * Gibt den Abstand der Ergebnisganglinie zur Referenzganglinie zurück.
         *
         * @return der Abstand.
         */
        public double getAbstand() {
            return abstand;
        }

        /**
         * Gibt die Ergebnisganglinie zurück.
         *
         * @return die Ergebnisganglinie.
         */
        public DoubleGanglinie getGanglinie() {
            return ganglinie.erzeugeKopie();
        }

        /**
         * Gibt den Index der ausgewählten Ganglinie in der Vergleichsliste zurück.
         *
         * @return der Index.
         */
        public int getIndex() {
            return index;
        }

        /**
         *
         * Gibt der Offset um den die Ergebnisganglinie verschoben wurde zurück.
         *
         * @return der Offset der Ergebnisganglinie.
         */
        public long getOffset() {
            return offset;
        }

        @Override
        public String toString() {
            String s;

            s = getClass() + "[";
            s += "index=" + index;
            s += ", abstand=" + abstand;
            s += ", offset=" + Duration.ofMillis(offset);
            s += ", ganglinie=" + ganglinie;
            s += "]";

            return s;
        }

        @Override
        public int hashCode() {
            return Objects.hash(abstand, ganglinie, index, offset);
        }

        @Override
        public boolean equals(Object obj) {
            if (this == obj) {
                return true;
            }
            if (obj == null) {
                return false;
            }
            if (!(obj instanceof PatternMatchingErgebnis)) {
                return false;
            }
            PatternMatchingErgebnis other = (PatternMatchingErgebnis) obj;
            return Double.doubleToLongBits(abstand) == Double.doubleToLongBits(other.abstand)
                    && Objects.equals(ganglinie, other.ganglinie) && index == other.index && offset == other.offset;
        }

    }

    /**
     * Konstruktor verstecken.
     */
    private GanglinienOperationen() {
        // nichts
    }

    /**
     * Addiert zwei Ganglinien, indem die Werte der vervollständigten
     * Stützstellenmenge addiert werden. Die beiden Ganglinien werden dabei nicht
     * verändert.
     *
     * @param g1 Erste Ganglinie
     * @param g2 Zweite Ganglinie
     * @return Die "Summe" der beiden Ganglinien
     */
    public static DoubleGanglinie addiere(final DoubleGanglinie g1, final DoubleGanglinie g2) {

        if ((((g1.getApproximation() != null) && (g2.getApproximation() != null))
                && !g1.getApproximation().getClass().equals(g2.getApproximation().getClass()))
                || ((g1.getApproximation() == null) ^ (g2.getApproximation() == null))) {
            throw new IllegalArgumentException(
                    "Die Operation kann nicht durchgeführt werden, da die beiden Ganglinien unterschiedliche Approximationsverfahren verwenden.");
        }

        Polyline p1 = new Polyline();
        p1.setStuetzstellen(g1.getStuetzstellen());
        p1.initialisiere();

        Polyline p2 = new Polyline();
        p2.setStuetzstellen(g2.getStuetzstellen());
        p2.initialisiere();

        DoubleGanglinie g = new DoubleGanglinie();
        Queue<Long> zeitstempel = vervollstaendigeStuetzstellen(g1, g2);

        for (final long z : zeitstempel) {
            if (g1.isValid(z) && g2.isValid(z)) {
                // irgendwas + irgendwas = irgendwas
                g.aktualisierePunkt(z, p1.get(z).getWert() + p2.get(z).getWert());
            } else if (g1.isValid(z) && (p2.getStuetzstellen().size() > 0)) {

                Stuetzstelle<Double> erste = p2.getStuetzstellen().get(0);
                Stuetzstelle<Double> letzte = p2.getStuetzstellen().get(p2.getStuetzstellen().size() - 1);

                if (z < erste.getZeitstempel()) {
                    // irgendwas + undefiniert = irgendwas, am Anfang
                    g.aktualisierePunkt(z, p1.get(z).getWert() + erste.getWert());
                } else if (z > letzte.getZeitstempel()) {
                    // irgendwas + undefiniert = irgendwas, am Ende
                    g.aktualisierePunkt(z, p1.get(z).getWert() + letzte.getWert());
                } else {
                    // irgendwas + undefiniert = undefiniert, in der Mitte
                    // g.put(z, null);
                    g.aktualisierePunkt(z, p1.get(z).getWert());
                }
            } else if (g2.isValid(z) && (p1.getStuetzstellen().size() > 0)) {

                Stuetzstelle<Double> erste = p1.getStuetzstellen().get(0);
                Stuetzstelle<Double> letzte = p1.getStuetzstellen().get(p1.getStuetzstellen().size() - 1);

                if (z < erste.getZeitstempel()) {
                    // undefiniert + irgendwas = irgendwas, am Anfang
                    g.aktualisierePunkt(z, erste.getWert() + p2.get(z).getWert());
                } else if (z > letzte.getZeitstempel()) {
                    // undefiniert + irgendwas = irgendwas, am Ende
                    g.aktualisierePunkt(z, letzte.getWert() + p2.get(z).getWert());
                } else {
                    // undefiniert + irgendwas = undefiniert, in der Mitte
                    // g.put(z, null);
                    g.aktualisierePunkt(z, p2.get(z).getWert());
                }
            } else {
                // undefiniert + irgendwas = undefiniert
                g.aktualisierePunkt(z, null);
            }
        }

        return g;
    }

    /**
     * Schneidet ein Intervall aus einer Ganglinie heraus. Existieren keine
     * Stützstellen in den Intervallgrenzen, werden an diesen Stellen mittels
     * Approximation durch Polyline Stützstellen hinzugefügt.
     * <p>
     * <em>Hinweis:</em> Das Intervall wird aus der Ganglinie im Parameter
     * ausgeschnitten.
     *
     * @param g Eine Ganglinie
     * @param i Auszuschneidendes Intervall
     * @return Der Intervallausschnitt
     */
    public static DoubleGanglinie auschneiden(final DoubleGanglinie g, final Intervall i) {
        Polyline p;

        p = new Polyline();
        p.setStuetzstellen(g.getStuetzstellen());
        p.initialisiere();

        // Stützstellen an den beiden Schnittpunkten ergänzen, falls nötig
        if (!g.hatZeitpunkt(i.getStart())) {
            g.setStuetzstelle(p.get(i.getStart()));
        }
        if (!g.hatZeitpunkt(i.getEnd())) {
            g.setStuetzstelle(p.get(i.getEnd()));
        }

        // Stützstellen außerhalb des Intervalls entfernen
        Iterator<Entry<Long, Double>> iterator;
        iterator = g.getWerte().entrySet().iterator();
        while (iterator.hasNext()) {
            long t;

            t = iterator.next().getKey();
            if (!i.contains(t)) {
                iterator.remove();
            }
        }

        return g;
    }

    /**
     * Berechnet den Abstand zweier Ganglinien mit Hilfe des Basisabstandsverfahren.
     * Es wir der Abstand anhand der vervollständigten Stützstellenmenge bestimmt.
     *
     * @param g1 Erste Ganglinie
     * @param g2 Zweite Ganglinie
     * @return Abstand nach dem Basisabstandsverfahren in Prozent.
     */
    public static double basisabstand(final DoubleGanglinie g1, final DoubleGanglinie g2) {

        DoubleGanglinie p1 = g1.erzeugeKopie();
        p1.setApproximation(new Polyline());

        DoubleGanglinie p2 = g2.erzeugeKopie();
        p2.setApproximation(new Polyline());

        final Queue<Long> zeitstempel = vervollstaendigeStuetzstellen(g1, g2);

        double fehler = fehler(p1, p2, zeitstempel);
        if (Double.isFinite(fehler)) {
            return fehler;
        }
        return Double.MAX_VALUE;
    }

    /**
     * Division zweier Ganglinien, indem die Werte der vervollständigten
     * Stützstellenmenge dividiert werden. Die beiden Ganglinien werden dabei nicht
     * verändert.
     *
     * @param g1 Erste Ganglinie
     * @param g2 Zweite Ganglinie
     * @return Das "Produkt" der beiden Ganglinien
     */
    public static DoubleGanglinie dividiere(final DoubleGanglinie g1, final DoubleGanglinie g2) {

        if ((((g1.getApproximation() != null) && (g2.getApproximation() != null))
                && !g1.getApproximation().getClass().equals(g2.getApproximation().getClass()))
                || ((g1.getApproximation() == null) ^ (g2.getApproximation() == null))) {
            throw new IllegalArgumentException(
                    "Die Operation kann nicht durchgeführt werden, da die beiden Ganglinien unterschiedliche Approximationsverfahren verwenden.");
        }

        Polyline p1 = new Polyline();
        p1.setStuetzstellen(g1.getStuetzstellen());
        p1.initialisiere();

        Polyline p2 = new Polyline();
        p2.setStuetzstellen(g2.getStuetzstellen());
        p2.initialisiere();

        DoubleGanglinie g = new DoubleGanglinie();
        Queue<Long> zeitstempel = vervollstaendigeStuetzstellen(g1, g2);

        for (final long z : zeitstempel) {

            Double a = null;
            Double b = null;

            if (g1.isValid(z) && g2.isValid(z)) {
                // irgendwas / irgendwas = irgendwas
                a = p1.get(z).getWert();
                b = p2.get(z).getWert();
            } else if (g1.isValid(z) && (p2.getStuetzstellen().size() > 0)) {

                Stuetzstelle<Double> erste = p2.getStuetzstellen().get(0);
                Stuetzstelle<Double> letzte = p2.getStuetzstellen().get(p2.getStuetzstellen().size() - 1);

                if (z < erste.getZeitstempel()) {
                    // irgendwas / undefiniert = irgendwas, am Anfang
                    a = p1.get(z).getWert();
                    b = erste.getWert();
                } else if (z > letzte.getZeitstempel()) {
                    // irgendwas / undefiniert = irgendwas, am Ende
                    a = p1.get(z).getWert();
                    b = letzte.getWert();
                } else {
                    // irgendwas / undefiniert = undefiniert, in der Mitte
                    a = p1.get(z).getWert();
                    b = 1.0;
                }
            } else if (g2.isValid(z) && (p1.getStuetzstellen().size() > 0)) {

                Stuetzstelle<Double> erste = p1.getStuetzstellen().get(0);
                Stuetzstelle<Double> letzte = p1.getStuetzstellen().get(p1.getStuetzstellen().size() - 1);

                if (z < erste.getZeitstempel()) {
                    // undefiniert / irgendwas = irgendwas, am Anfang
                    a = erste.getWert();
                    b = p2.get(z).getWert();
                } else if (z > letzte.getZeitstempel()) {
                    // undefiniert + irgendwas = irgendwas, am Ende
                    a = letzte.getWert();
                    b = p2.get(z).getWert();
                } else {
                    // undefiniert + irgendwas = undefiniert, in der Mitte
                    a = p2.get(z).getWert();
                    b = 1.0;
                }
            }

            if ((a != null) && (b != null)) {
                Double c = a / b;
                if (Double.isNaN(c) || Double.isInfinite(c)) {
                    g.aktualisierePunkt(z, null);
                } else {
                    g.aktualisierePunkt(z, c);
                }
            } else {
                g.aktualisierePunkt(z, null);
            }
        }

        return g;
    }

    /**
     * Bestimmt den prozentualen Fehler (Abstand) zweier Ganglinien anhand gegebener
     * Messpunkte.
     *
     * @param g1          die erste Ganglinie.
     * @param g2          die zweite Ganglinie.
     * @param zeitstempel die "Messpunkte"
     * @return der prozentuale Fehler.
     */
    private static double fehler(final DoubleGanglinie g1, final DoubleGanglinie g2, final Queue<Long> zeitstempel) {

        // Quadratischen Fehler bestimmen
        double summe = 0;
        int undefinierte = 0;
        for (final long z : zeitstempel) {
            if ((g1.getStuetzstelle(z).getWert() != null) && (g2.getStuetzstelle(z).getWert() != null)) {
                double x;

                x = g2.getStuetzstelle(z).getWert() - g1.getStuetzstelle(z).getWert();
                summe += x * x;
            } else {
                // Undefinierte Stützstellen werden nicht berücksichtigt
                undefinierte++;
            }
        }

        if ((zeitstempel.size() - undefinierte) == 0) {
            // Die beiden Ganglinienintevalle überschneiden sich nicht.
            return Double.POSITIVE_INFINITY;
        }
        double fehler = Math.sqrt(summe / (zeitstempel.size() - undefinierte));

        // Prozentualen Fehler bestimmen
        summe = 0;
        for (final long z : zeitstempel) {
            if ((g1.getStuetzstelle(z).getWert() != null) && (g2.getStuetzstelle(z).getWert() != null)) {
                double x;

                x = (g1.getStuetzstelle(z).getWert() + g2.getStuetzstelle(z).getWert()) / 2;
                summe += x * x;
            }
        }
        return (fehler * 100) / (Math.sqrt(summe / (zeitstempel.size() - undefinierte)));
    }

    /**
     * Konvertiert eine Ganglinie in einen lesbaren Text. Dazu wird zwischen jeder
     * Stützstelle einfach ein Zeilenbruch eingefügt. Enthält die Ganglinie keine
     * Stützstellen wird eine kurze Notiz ausgegeben.
     *
     * @param g eine Ganglinie.
     * @return ein String der jede Stützstelle auf eine eigene Zeile schreibt.
     */
    public static String formatierterText(final DoubleGanglinie g) {
        StringBuilder txt = new StringBuilder();

        txt.append("Intervall: " + g.getIntervall());
        txt.append("\nApproximation: " + g.getApproximation());
        for (final Stuetzstelle<?> s : g.getStuetzstellen()) {
            txt.append("\n" + s);
        }

        if (g.punkteAnzahl() == 0) {
            txt.append("\nKeine Stützstellen vorhanden.");
        }

        return txt.toString();
    }

    /**
     * Berechnet den Abstand zweier Ganglinien mit Hilfe des komplexen
     * Abstandsverfahren. Es werden jeweils die Werte an den Intervallgrenzen
     * verglichen.
     *
     * @param g1         Erste Ganglinie.
     * @param g2         Zweite Ganglinie.
     * @param intervalle die Anzahl der zu vergleichenden Intervalle.
     * @return Abstand nach dem komplexen Abstandsverfahren in Prozent.
     */
    public static double komplexerAbstand(final DoubleGanglinie g1, final DoubleGanglinie g2, final int intervalle) {

        // Zu betrachtendes Intervall und Intervallbreite bestimmen
        long start;
        if (g1.ersterZeitpunkt() < g2.ersterZeitpunkt()) {
            start = g2.ersterZeitpunkt();
        } else {
            start = g1.ersterZeitpunkt();
        }

        long ende;
        if (g1.letzterZeitpunkt() < g2.letzterZeitpunkt()) {
            ende = g1.letzterZeitpunkt();
        } else {
            ende = g2.letzterZeitpunkt();
        }
        long breite = Math.round((double) (ende - start) / intervalle);

        return komplexerAbstand(g1, g2, breite);
    }

    /**
     * Berechnet den Abstand zweier Ganglinien mit Hilfe des komplexen
     * Abstandsverfahren. Es werden jeweils die Werte an den Intervallgrenzen
     * verglichen.
     *
     * @param g1              Erste Ganglinie.
     * @param g2              Zweite Ganglinie.
     * @param intervallBreite die Breite der zu vergleichenden Intervalle.
     * @return Abstand nach dem komplexen Abstandsverfahren in Prozent.
     */
    public static double komplexerAbstand(final DoubleGanglinie g1, final DoubleGanglinie g2,
            final long intervallBreite) {

        // Zu betrachtendes Intervall und Intervallbreite bestimmen
        long start;
        if (g1.ersterZeitpunkt() < g2.ersterZeitpunkt()) {
            start = g2.ersterZeitpunkt();
        } else {
            start = g1.ersterZeitpunkt();
        }

        long ende;
        if (g1.letzterZeitpunkt() < g2.letzterZeitpunkt()) {
            ende = g1.letzterZeitpunkt();
        } else {
            ende = g2.letzterZeitpunkt();
        }

        if (start > ende) {
            // Die beiden Ganglinienintervalle überschneiden sich nicht.
            return Double.MAX_VALUE;
        }

        // Haltepunkte bestimmen
        final Queue<Long> zeitstempel = new LinkedList<>();
        for (long i = start; i < ende; i += intervallBreite) {
            zeitstempel.add(i);
        }
        zeitstempel.add(ende);

        return fehler(g1, g2, zeitstempel);
    }

    /**
     * Multiplikation zweier Ganglinien, indem die Werte der vervollständigten
     * Stützstellenmenge multipliziert werden. Die beiden Ganglinien werden dabei
     * nicht verändert.
     *
     * @param g1 Erste Ganglinie
     * @param g2 Zweite Ganglinie
     * @return Das "Produkt" der beiden Ganglinien
     */
    public static DoubleGanglinie multipliziere(final DoubleGanglinie g1, final DoubleGanglinie g2) {

        if ((((g1.getApproximation() != null) && (g2.getApproximation() != null))
                && !g1.getApproximation().getClass().equals(g2.getApproximation().getClass()))
                || ((g1.getApproximation() == null) ^ (g2.getApproximation() == null))) {
            throw new IllegalArgumentException(
                    "Die Operation kann nicht durchgeführt werden, da die beiden Ganglinien unterschiedliche Approximationsverfahren verwenden.");
        }

        Polyline p1 = new Polyline();
        p1.setStuetzstellen(g1.getStuetzstellen());
        p1.initialisiere();

        Polyline p2 = new Polyline();
        p2.setStuetzstellen(g2.getStuetzstellen());
        p2.initialisiere();

        DoubleGanglinie g = new DoubleGanglinie();
        Queue<Long> zeitstempel = vervollstaendigeStuetzstellen(g1, g2);

        for (final long z : zeitstempel) {
            if (g1.isValid(z) && g2.isValid(z)) {
                // irgendwas * irgendwas = irgendwas
                g.aktualisierePunkt(z, p1.get(z).getWert() * p2.get(z).getWert());
            } else if (g1.isValid(z) && (p2.getStuetzstellen().size() > 0)) {

                Stuetzstelle<Double> erste = p2.getStuetzstellen().get(0);
                Stuetzstelle<Double> letzte = p2.getStuetzstellen().get(p2.getStuetzstellen().size() - 1);

                if (z < erste.getZeitstempel()) {
                    // irgendwas * undefiniert = irgendwas, am Anfang
                    g.aktualisierePunkt(z, p1.get(z).getWert() * erste.getWert());
                } else if (z > letzte.getZeitstempel()) {
                    // irgendwas * undefiniert = irgendwas, am Ende
                    g.aktualisierePunkt(z, p1.get(z).getWert() * letzte.getWert());
                } else {
                    // irgendwas * undefiniert = undefiniert, in der Mitte
                    g.aktualisierePunkt(z, p1.get(z).getWert());
                }
            } else if (g2.isValid(z) && (p1.getStuetzstellen().size() > 0)) {

                Stuetzstelle<Double> erste = p1.getStuetzstellen().get(0);
                Stuetzstelle<Double> letzte = p1.getStuetzstellen().get(p1.getStuetzstellen().size() - 1);

                if (z < erste.getZeitstempel()) {
                    // undefiniert * irgendwas = irgendwas, am Anfang
                    g.aktualisierePunkt(z, erste.getWert() * p2.get(z).getWert());
                } else if (z > letzte.getZeitstempel()) {
                    // undefiniert * irgendwas = irgendwas, am Ende
                    g.aktualisierePunkt(z, letzte.getWert() * p2.get(z).getWert());
                } else {
                    // undefiniert * irgendwas = undefiniert, in der Mitte
                    g.aktualisierePunkt(z, p2.get(z).getWert());
                }
            } else {
                // undefiniert * irgendwas = undefiniert
                g.aktualisierePunkt(z, null);
            }
        }

        return g;
    }

    /**
     * Normiert die Stützstellen einer Ganglinie. Mehrere hintereinander folgende
     * undefinierte Stützstellen werden zusammengefasst. Der Abstand der
     * Stützstellen wird auf ein definiertes Intervall normiert. Der Intervallzyklus
     * beginnt mit dem Zeitstempel der ersten Stützstelle.
     * <p>
     * <em>Hinweis:</em> die Ganglinie im Parameter wird verändert.
     *
     * @param g       eine Ganglinie.
     * @param abstand der gewünschte Stützstellenabstand.
     * @return die normierte Ganglinie.
     */
    public static DoubleGanglinie normiere(final DoubleGanglinie g, final long abstand) {
        final Polyline p;
        final SortedMap<Long, Double> stuetzstellen;
        final long halberAbstand;

        p = new Polyline();
        p.setStuetzstellen(g.getStuetzstellen());
        p.initialisiere();

        stuetzstellen = new TreeMap<>();
        halberAbstand = abstand / 2;
        stuetzstellen.put(g.ersterZeitpunkt(), g.getWert(g.ersterZeitpunkt()));
        for (long i = g.ersterZeitpunkt() + halberAbstand; i < g.letzterZeitpunkt(); i += abstand) {
            Intervall intervall = Intervall.of(i - halberAbstand, i + halberAbstand);
            if (g.isValid(intervall)) {
                stuetzstellen.put(i, p.integral(intervall) / intervall.getLength());
            } else {
                stuetzstellen.put(i, null);
            }
        }
        stuetzstellen.put(g.letzterZeitpunkt(), g.getWert(g.letzterZeitpunkt()));

        g.loeschen();
        g.aktualisierePunkte(stuetzstellen);
        return g;
    }

    /**
     * Führt das Pattern-Matching einer Menge von Ganglinien mit einer
     * Referenzganglinie aus. Das Ergebnis ist die Ganglinie aus der Menge mit dem
     * geringsten Abstand zur Referenzganglinie.
     * <p>
     * Zusätzlich zu der Liste von Vergleichsganglinien wird jede dieser Ganglinien
     * im angegebenen Offset in {@code intervall} Schritten verschoben. Jede dieser
     * so entstanden Ganglinien wird ebenfalls mit der Referenzganglinie verglichen.
     * <p>
     * Ergebnis ist die vorgegebene oder erzeugte Ganglinie mit dem geringsten
     * Abstand. Der Index im Ergebnis ist der Index in der Ganglinienliste, aus der
     * die Ergebnisganglinie hervorgegangen ist.
     *
     * @param referenz        die Referenzganglinie.
     * @param vergleichsListe die Liste von zu vergleichenden Ganglinien.
     * @param offsetVor       der Offset, in dem die Ganglinien nach vorn verschoben
     *                        werden kann.
     * @param offsetNach      der Offset, in dem die Ganglinien nach hinten
     *                        verschoben werden kann.
     * @param schrittweite    das Intervall, in dem die Ganglinien innerhalb des
     *                        Offsets verschoben werden.
     * @return das Ergebnis des Pattern-Matching.
     */
    public static PatternMatchingErgebnis patternMatching(final DoubleGanglinie referenz,
            final List<DoubleGanglinie> vergleichsListe, final long offsetVor, final long offsetNach,
            final long schrittweite) {
        if (referenz == null) {
            throw new IllegalArgumentException(MessageFormat.format("Argument darf nicht null sein: {0}", "referenz"));
        }
        if (vergleichsListe.isEmpty()) {
            throw new IllegalArgumentException(
                    MessageFormat.format("Argument darf keine leere Liste sein: {0}", "vergleichsListe"));
        }
        if (offsetNach < 0) {
            throw new IllegalArgumentException(MessageFormat.format(
                    "IllegalArgumentExceptionGroesserGleichNull = Argument muss eine Zahl größer oder gleich 0 sein: {0}",
                    "offsetNach"));
        }
        if (offsetVor < 0) {
            throw new IllegalArgumentException(MessageFormat.format(
                    "IllegalArgumentExceptionGroesserGleichNull = Argument muss eine Zahl größer oder gleich 0 sein: {0}",
                    "offsetVor"));
        }
        if (schrittweite <= 0) {
            throw new IllegalArgumentException(
                    MessageFormat.format("Argument muss eine Zahl größer 0 sein: {0}", "schrittweite"));
        }

        PatternMatchingErgebnis ergebnis = null;
        for (int i = 0; i < vergleichsListe.size(); i++) {
            for (long offset = -offsetVor; offset <= offsetNach; offset += schrittweite) {
                DoubleGanglinie g = vergleichsListe.get(i).erzeugeKopie();
                verschiebe(g, offset);
                double abstand = basisabstand(referenz, g);
                PatternMatchingErgebnis matchingErgebnis = new PatternMatchingErgebnis(g.erzeugeKopie(), i, abstand,
                        offset);
                if (ergebnis == null || ergebnis.abstand > matchingErgebnis.abstand) {
                    ergebnis = matchingErgebnis;
                }
            }
        }

        return ergebnis;
    }

    /**
     * Subtraktion zweier Ganglinien, indem die Werte der vervollständigten
     * Stützstellenmenge subtrahiert werden. Die beiden Ganglinien werden dabei
     * nicht verändert.
     *
     * @param g1 Erste Ganglinie
     * @param g2 Zweite Ganglinie
     * @return Die "Differenz" der beiden Ganglinien
     */
    public static DoubleGanglinie subtrahiere(final DoubleGanglinie g1, final DoubleGanglinie g2) {

        if ((((g1.getApproximation() != null) && (g2.getApproximation() != null))
                && !g1.getApproximation().getClass().equals(g2.getApproximation().getClass()))
                || ((g1.getApproximation() == null) ^ (g2.getApproximation() == null))) {
            throw new IllegalArgumentException(
                    "Die Operation kann nicht durchgeführt werden, da die beiden Ganglinien unterschiedliche Approximationsverfahren verwenden.");
        }

        Polyline p1 = new Polyline();
        p1.setStuetzstellen(g1.getStuetzstellen());
        p1.initialisiere();

        Polyline p2 = new Polyline();
        p2.setStuetzstellen(g2.getStuetzstellen());
        p2.initialisiere();

        DoubleGanglinie g = new DoubleGanglinie();
        Queue<Long> zeitstempel = vervollstaendigeStuetzstellen(g1, g2);

        for (final long z : zeitstempel) {
            if (g1.isValid(z) && g2.isValid(z)) {
                // irgendwas - irgendwas = irgendwas
                g.aktualisierePunkt(z, p1.get(z).getWert() - p2.get(z).getWert());
            } else if (g1.isValid(z) && (p2.getStuetzstellen().size() > 0)) {

                Stuetzstelle<Double> erste = p2.getStuetzstellen().get(0);
                Stuetzstelle<Double> letzte = p2.getStuetzstellen().get(p2.getStuetzstellen().size() - 1);

                if (z < erste.getZeitstempel()) {
                    // irgendwas - undefiniert = irgendwas, am Anfang
                    g.aktualisierePunkt(z, p1.get(z).getWert() - erste.getWert());
                } else if (z > letzte.getZeitstempel()) {
                    // irgendwas - undefiniert = irgendwas, am Ende
                    g.aktualisierePunkt(z, p1.get(z).getWert() - letzte.getWert());
                } else {
                    // irgendwas + undefiniert = undefiniert, in der Mitte
                    g.aktualisierePunkt(z, p1.get(z).getWert());
                }
            } else if (g2.isValid(z) && (p1.getStuetzstellen().size() > 0)) {

                Stuetzstelle<Double> erste = p1.getStuetzstellen().get(0);
                Stuetzstelle<Double> letzte = p1.getStuetzstellen().get(p1.getStuetzstellen().size() - 1);

                if (z < erste.getZeitstempel()) {
                    // undefiniert - irgendwas = irgendwas, am Anfang
                    g.aktualisierePunkt(z, erste.getWert() - p2.get(z).getWert());
                } else if (z > letzte.getZeitstempel()) {
                    // undefiniert - irgendwas = irgendwas, am Ende
                    g.aktualisierePunkt(z, letzte.getWert() - p2.get(z).getWert());
                } else {
                    // undefiniert - irgendwas = undefiniert, in der Mitte
                    g.aktualisierePunkt(z, p2.get(z).getWert());
                }
            } else {
                // undefiniert - irgendwas = undefiniert
                g.aktualisierePunkt(z, null);
            }
        }

        return g;
    }

    /**
     * Verbindet zwei Ganglinien durch Konkatenation. Es werden die Stützstellen
     * beider Ganglinien zu einer neuen Ganglinien zusammengefasst. Dies ist nur
     * möglich, wenn sich die Stützstellenmengen nicht überschneiden. Berühren sich
     * die beiden Ganglinien wird im Berührungspunkt er Mittelwert der beiden
     * Stützstellen gebildet.
     * <p>
     * Wird der maximale Abstand der beiden Ganglinien überschritten, ist der
     * Bereich zwischen den Ganglinien undefiniert.
     *
     * @param g1         Erste Ganglinie
     * @param g2         Zweite Ganglinie
     * @param maxAbstand der maximale Abstand der beiden Ganglinien.
     * @return Konkatenation der beiden Ganglinien
     */
    public static DoubleGanglinie verbinde(final DoubleGanglinie g1, final DoubleGanglinie g2, final long maxAbstand) {
        if (g1.getIntervall().intersect(g2.getIntervall())) {
            throw new IllegalArgumentException("Die zu verbindenden Ganglinien dürfen sich nicht überschneiden.");
        }

        final DoubleGanglinie g = new DoubleGanglinie();
        g.aktualisierePunkte(g1.getWerte());
        for (final long t : g2.getZeitpunkte()) {
            if (g.hatZeitpunkt(t)) {

                Double d1 = g.getWert(t);
                Double d2 = g2.getWert(t);
                if ((d1 != null) && (d2 != null)) {
                    g.aktualisierePunkt(t, (d1 + d2) / 2);
                } else {
                    g.aktualisierePunkt(t, null);
                }
            } else {
                g.aktualisierePunkt(t, g2.getWert(t));
            }
        }

        if (g1.letzterZeitpunkt() < g2.ersterZeitpunkt()) {
            // g1 liegt vor g2

            final long abstand;

            abstand = g2.ersterZeitpunkt() - g1.letzterZeitpunkt();
            if (abstand > maxAbstand) {
                g.aktualisierePunkt(g1.letzterZeitpunkt() + (abstand / 2), null);
            }
        } else {
            // g2 liegt vor g1

            final long abstand;

            abstand = g1.ersterZeitpunkt() - g2.letzterZeitpunkt();
            if (abstand > maxAbstand) {
                g.aktualisierePunkt(g2.letzterZeitpunkt() + (abstand / 2), null);
            }
        }

        return g;
    }

    /**
     * Verschiebt eine Ganglinie auf der Zeitachse.
     * <p>
     * <em>Hinweis:</em> Es wird die Ganglinie im Parameter verschoben.
     *
     * @param g      Zu verschiebende Ganglinie
     * @param offset Offset um den die Ganglinie verschoben werden soll
     * @return Die verschobene Ganglinie
     */
    public static DoubleGanglinie verschiebe(final DoubleGanglinie g, final long offset) {
        SortedMap<Long, Double> stuetzstellen;

        stuetzstellen = new TreeMap<>();
        for (final long t : g.getZeitpunkte()) {
            stuetzstellen.put(t + offset, g.getWert(t));
        }
        g.loeschen();
        g.aktualisierePunkte(stuetzstellen);
        return g;
    }

    /**
     * Verschiebt eine Ganglinie auf der Zeitachse um ein halbes
     * Stützstellenintervall. Jede Stützstelle wird um den halben Abstand zur
     * nächsten Stützstelle verschoben. Die letzte Stützstelle wird um den halben
     * Abstand zur vorherigen Stützstelle verschoben. Gibt es nur eine Stützstelle,
     * wird diese nicht verschoben.
     * <p>
     * <em>Hinweis:</em> Es wird die Ganglinie im Parameter verschoben.
     *
     * @param g Zu verschiebende Ganglinie
     * @return Die verschobene Ganglinie
     */
    public static DoubleGanglinie verschiebeUmHalbesIntervall(final DoubleGanglinie g) {

        if (g.punkteAnzahl() < 2) {
            throw new IllegalArgumentException("Die Ganglinie muss mindestens zwei Stützstellen besitzen.");
        }

        long t0 = Long.MIN_VALUE;
        long intervall = 0;
        SortedMap<Long, Double> stuetzstellen = new TreeMap<>();
        for (final Long t : g.getZeitpunkte()) {
            if (t0 == Long.MIN_VALUE) {
                t0 = t;
            } else {
                intervall = t - t0;
                stuetzstellen.put(t0 + (intervall / 2), g.getWert(t0));
                t0 = t;
            }
        }
        stuetzstellen.put(g.letzterZeitpunkt() + (intervall / 2), g.getWert(g.letzterZeitpunkt()));
        g.loeschen();
        g.aktualisierePunkte(stuetzstellen);
        return g;
    }

    /**
     * Verschmilzt eine Ganglinie mit einer anderen. Dabei wird das gewichtete
     * arithmetische Mittel der vervollständigten Stützstellen gebildet. Die zweite
     * Ganglinie hat immer das Gewicht 1.
     *
     * @param g1      die Ganglinie mit der verschmolzen wird. Sie hat immer das
     *                Gewicht 1.
     * @param g2      die Ganglinie die verschmolzen wird.
     * @param gewicht das Gewicht der zweiten Ganglinie.
     * @return das Ergebnis der Verschmelzung.
     */
    public static DoubleGanglinie verschmelze(final DoubleGanglinie g1, final DoubleGanglinie g2, final long gewicht) {

        Polyline p1 = new Polyline();
        p1.setStuetzstellen(g1.getStuetzstellen());
        p1.initialisiere();

        Polyline p2 = new Polyline();
        p2.setStuetzstellen(g2.getStuetzstellen());
        p2.initialisiere();

        DoubleGanglinie g = new DoubleGanglinie();
        Queue<Long> zeitstempel = vervollstaendigeStuetzstellen(g1, g2);

        while (!zeitstempel.isEmpty()) {
            long z;

            z = zeitstempel.poll();

            if (g1.isValid(z) && g2.isValid(z)) {
                g.aktualisierePunkt(z, (p1.get(z).getWert() + (p2.get(z).getWert() * gewicht)) / (gewicht + 1));
            } else {
                // undefiniert # irgendwas = irgendwas
                if (g1.isValid(z)) {
                    g.aktualisierePunkt(z, p1.get(z).getWert());
                } else if (g2.isValid(z)) {
                    g.aktualisierePunkt(z, p2.get(z).getWert());
                } else {
                    // beide Operanden sind undefiniert
                    g.aktualisierePunkt(z, null);
                }
            }
        }

        return g;
    }

    /**
     * Bestimmt die vereinigte Menge der Stützstellen beider Ganglinien.
     *
     * @param g1 Erste Ganglinie
     * @param g2 Zweite Ganglinie
     * @return Menge von Stützstellenreferenzen in Form von Zeitstempeln
     */
    private static LinkedList<Long> vervollstaendigeStuetzstellen(final DoubleGanglinie g1, final DoubleGanglinie g2) {
        SortedSet<Long> zeitstempel;

        zeitstempel = new TreeSet<>();
        zeitstempel.addAll(g1.getZeitpunkte());
        zeitstempel.addAll(g2.getZeitpunkte());

        return new LinkedList<>(zeitstempel);
    }
}
