/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.5 Funktionen Ganglinie
 * Copyright (C) 2011-2015 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.gllib.gllib.shared;

import java.time.LocalDateTime;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;

/**
 * Ein einfacher Scheduler für {@link CronJob}s. "Einfach" heißt, dass beim
 * ausführen von Aktionen nicht geprüft wird, ob ein Cron-Job noch läuft. Ein
 * einmal gestarteter Cron-Job wird "vergessen" und darauf vertraut, dass er
 * sich rechtzeitig vor dem nächsten Zyklus beendet und überhaupt irgendwann
 * beendet.
 * 
 * @author BitCtrl Systems GmbH, Falko Schumann
 */
public class CronScheduler {

    private final Map<Long, CronJob> jobs = new HashMap<>();
    
    private ScheduledExecutorService executor = Executors.newSingleThreadScheduledExecutor();
    private ScheduledFuture<?> scheduled;


    /**
     * Nimmt einen neuen Cron-Job in die Jobliste auf. Existiert bereits ein Job mit
     * der selben Job-ID, dann wird dieser überschrieben.
     * 
     * @param job ein Cron-Job
     * @return die Id des geplanten Cron-Jobs (entspricht {@link CronJob#getId()}).
     */
    public long schedule(final CronJob job) {
        synchronized (jobs) {
            jobs.put(job.getId(), job);
        }
        return job.getId();
    }

    /**
     * Entfernt einen Job aus der Jobliste. Existiert kein Job mit der ID, passiert
     * nichts.
     * 
     * @param jobId eine Job-ID.
     */
    public void deschedule(final long jobId) {
        synchronized (jobs) {
            jobs.remove(jobId);
        }
    }

    /**
     * Gibt den Cron-Job mit einer bestimmten ID zurück. Existiert unter der ID kein
     * Job, wird {@code null} zurückgegeben.
     * 
     * @param jobId eine Job-ID.
     * @return der Cron-Job oder {@code null}.
     */
    public CronJob getCronJob(final long jobId) {
        synchronized (jobs) {
            return jobs.get(jobId);
        }
    }

    /**
     * Startet den Scheduler. Es wird zyklisch geprüft, welche Cron-Jobs ausgeführt
     * werden müssen. Auszuführende Jobs werden entsprechend jeweils in einem
     * eigenen Thread gestartet.
     */
    public final synchronized void start() {
        if (scheduled != null) {
            throw new IllegalStateException("Der Cron-Scheduler wurde bereits gestartet.");
        }

        scheduled = executor.scheduleAtFixedRate(() -> runJobs(), 0, 1, TimeUnit.MINUTES);
    }

    private void runJobs() {
        
        // Führe alle Cron-Jobs aus, die jetzt dran sind
        LocalDateTime now = LocalDateTime.now();
        for (final CronJob job : jobs.values()) {
            if (job.getPattern() != null && job.getPattern().match(now)) {
                new Thread(job, job.getName() + ", gestartet: " + now).start();
            }
        }
    }

    /**
     * Stopt den Scheduler. Laufende Cron-Jobs werden bis zu ihrem normalen Ende
     * ausgeführt. Es werden keine neuen Jobs mehr angestoßen.
     */
    public final synchronized void stop() {
        if (scheduled == null) {
            throw new IllegalStateException("Der Cron-Scheduler wurde nicht gestartet.");
        }

        scheduled.cancel(true);
        scheduled = null;
    }

    /**
     * Fragt, ob der Scheduler gestartet wurde.
     * 
     * @return {@code true}, wenn der Scheduler gestartet ist.
     */
    public final synchronized boolean isStarted() {
        return scheduled != null;
    }

    /**
     * Läßt den aktuellen Thread für die angegebene Zeit schlafen. Diese Methode
     * kann von abgeleiteten Klassen überschrieben werden um eigene Pausemethoden zu
     * implementieren. Die Standardimplementierung verwendet
     * {@link Thread#sleep(long)}.
     * 
     * @param millis die Länge der gewünschten Pause in Millisekunden.
     * @throws InterruptedException wenn das Sleep unterbrochen wurde.
     */
    public void sleep(final long millis) throws InterruptedException {
        Thread.sleep(millis);
    }

    /**
     * Gibt die aktuelle Zeit zurück. Diese Methode kann von abgeleiteten Klassen
     * überschrieben werden um eigene Zeitbestimmung zu implementieren. Die
     * Standardimplementierung verwendet {@link System#currentTimeMillis()}.
     * 
     * @return die aktuelle Zeit in Millisekunden.
     */
    public long getTime() {
        return System.currentTimeMillis();
    }
}
